from typing import List

from sqlalchemy.orm import Session
from sqlalchemy import select

from open_aoi.models import Base
from open_aoi.exceptions import IntegrityError


class Controller:
    _model: Base

    def __init__(self, session: Session):
        self.session = session

    def retrieve(self, id: int) -> Base:
        q = select(self._model).where(self._model.id == id)
        res = self.session.scalars(q).one_or_none()
        return res

    def delete(self, obj: Base):
        if self.allow_delete_hook(obj.id):
            self.session.query(self._model).filter(self._model.id == obj.id).delete()
            self.post_delete_hook(obj)
        else:
            raise IntegrityError(
                "Unable to delete. Object is a dependency for other objects."
            )

    def delete_by_id(self, id: int):
        if self.allow_delete_hook(id):
            obj = self.retrieve(id)
            self.session.query(self._model).filter(self._model.id == id).delete()
            self.post_delete_hook(obj)
        else:
            raise IntegrityError(
                "Unable to delete. Object is a dependency for other objects."
            )

    def commit(self):
        self.session.commit()

    def list(self) -> List[Base]:
        return self.session.query(self._model).all()

    def list_nested(self) -> List[Base]:
        raise NotImplementedError()

    def create(self, *args, **kwargs):
        raise NotImplementedError()

    def allow_delete_hook(self, id: int) -> bool:
        raise NotImplementedError()

    def post_delete_hook(self, obj: Base):
        pass



import os

MYSQL_DATABASE = os.environ["MYSQL_DATABASE"]
assert MYSQL_DATABASE

MYSQL_USER = os.environ["MYSQL_USER"]
assert MYSQL_USER

MYSQL_PASSWORD = os.environ["MYSQL_PASSWORD"]
assert MYSQL_PASSWORD

MYSQL_PORT = os.environ["MYSQL_PORT"]
assert MYSQL_PORT
try:
    MYSQL_PORT = int(MYSQL_PORT)
except ValueError:
    raise RuntimeError("Failed to parse mysql port from environment")

AOI_OPERATOR_INITIAL_PASSWORD = os.environ["AOI_OPERATOR_INITIAL_PASSWORD"]
assert AOI_OPERATOR_INITIAL_PASSWORD

AOI_ADMINISTRATOR_INITIAL_PASSWORD = os.environ["AOI_ADMINISTRATOR_INITIAL_PASSWORD"]
assert AOI_ADMINISTRATOR_INITIAL_PASSWORD

STORAGE_SECRET = os.environ["STORAGE_SECRET"]
assert STORAGE_SECRET and len(STORAGE_SECRET) > 10

MINIO_PORT = os.environ["MINIO_PORT"]
assert MINIO_PORT

MINIO_ROOT_USER = os.environ["MINIO_ROOT_USER"]
assert MINIO_ROOT_USER

MINIO_ROOT_PASSWORD = os.environ["MINIO_ROOT_PASSWORD"]
assert MINIO_ROOT_PASSWORD



from sqlalchemy import select

from open_aoi.controllers import Controller
from open_aoi.models import DefectTypeModel, ControlHandlerModel


class DefectTypeController(Controller):
    _model = DefectTypeModel

    def allow_delete_hook(self, id: int) -> bool:
        return not self.session.query(
            select(ControlHandlerModel)
            .where(ControlHandlerModel.defect_type_id == id)
            .exists()
        ).scalar()

    def create(self, title: str, description: str):
        obj = DefectTypeModel(
            title=title,
            description=description,
        )
        self.session.add(obj)
        return obj



from datetime import datetime

from open_aoi.controllers import Controller
from open_aoi.models import (
    InspectionModel,
    InspectionLogModel,
    ControlTargetModel,
)


class ControlLogController(Controller):
    _model = InspectionLogModel

    def create(
        self,
        control_target: ControlTargetModel,
        inspection: InspectionModel,
        log: str,
        passed: bool,
    ) -> InspectionLogModel:
        obj = InspectionLogModel(
            control_target=control_target,
            inspection=inspection,
            log=log,
            passed=passed,
            created_at=datetime.now(),
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        return False  # Do not delete log



from typing import List
from datetime import datetime

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from open_aoi.controllers import Controller
from open_aoi.models import (
    TemplateModel,
    ControlTargetModel,
    ControlZoneModel,
    AccessorModel,
)


class ControlZoneController(Controller):
    _model = ControlZoneModel

    def create(
        self,
        title: str,
        template: TemplateModel,
        accessor: AccessorModel,
    ) -> ControlZoneModel:
        obj = ControlZoneModel(
            title=title, template=template, created_by=accessor, created_at=datetime.now(), rotation=0
        )
        self.session.add(obj)
        return obj

    def list_nested(self) -> List[ControlZoneModel]:
        return (
            self.session.query(self._model)
            .options(joinedload(ControlZoneModel.cc))
            .options(joinedload(ControlZoneModel.control_target_list))
            .all()
        )

    def allow_delete_hook(self, id: int) -> bool:
        return not self.session.query(
            select(ControlTargetModel)
            .where(ControlTargetModel.control_zone_id == id)
            .exists()
        ).scalar()



from typing import List
import numpy as np
import cv2 as cv
from PIL import Image


def crop_stat_cv(im: np.ndarray, cv_stat_value: List[int]) -> np.ndarray:
    # Function parse CV connected component detection statics (values)
    # to cut out component from provided image
    t = cv_stat_value[cv.CC_STAT_TOP]
    l = cv_stat_value[cv.CC_STAT_LEFT]

    w = cv_stat_value[cv.CC_STAT_WIDTH]
    h = cv_stat_value[cv.CC_STAT_HEIGHT]

    return im[t : t + h, l : l + w]


def crop_stat_image(im: Image.Image, cv_stat_value: List[int]) -> Image.Image:
    # Wrapper for cropping PIL images
    im = np.array(im)
    im = crop_stat_cv(im, cv_stat_value)
    return Image.fromarray(im)



from dotenv import load_dotenv

assert load_dotenv(".env")

import unittest
import numpy as np
from PIL import Image

from open_aoi.models import engine
from open_aoi.controllers.template import TemplateController
from open_aoi.controllers.accessor import AccessorController
from open_aoi.controllers.control_handler import ControlHandlerController
from open_aoi.controllers.defect_type import DefectTypeController
from open_aoi.controllers.control_zone import ControlZoneController
from open_aoi.controllers.connected_component import ConnectedComponentController
from open_aoi.controllers.control_target import ControlTargetController

from sqlalchemy.orm import Session


class TemplateDatabaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.template_controller = TemplateController(self.session)
        self.accessor_controller = AccessorController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)

    def test_create_delete(self):
        template = self.template_controller.create("Test", self.accessor, None)
        self.template_controller.delete(template)

    def test_list_nested(self):
        self.template_controller.list_nested()


class TemplateMinioTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.template_controller = TemplateController(self.session)
        self.accessor_controller = AccessorController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)

        im = np.random.rand(100, 100, 3) * 255
        self.im = Image.fromarray(im.astype("uint8"))

        self.template = self.template_controller.create("Test", None, self.accessor)

    def test_create_destroy(self):
        self.template.publish_image(self.im)
        self.template.destroy_image()

    def test_materialize(self):
        self.template.publish_image(self.im)
        self.template.materialize_image()
        self.template.destroy_image()


class ControlHandlerDatabaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.control_handler_controller = ControlHandlerController(self.session)
        self.accessor_controller = AccessorController(self.session)
        self.defect_type_controller = DefectTypeController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)
        self.defect_type = self.defect_type_controller.retrieve(1)

    def test_create_delete(self):
        control_handler = self.control_handler_controller.create(
            "Test", "test", self.defect_type
        )
        self.control_handler_controller.delete(control_handler)

    def test_list_nested(self):
        self.control_handler_controller.list_nested()


class ControlHandlerMinioTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.control_handler_controller = ControlHandlerController(self.session)
        self.accessor_controller = AccessorController(self.session)
        self.defect_type_controller = DefectTypeController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)
        self.defect_type = self.defect_type_controller.retrieve(1)

        self.source = """
parameters = []
process = lambda: print('hello open-aoi!')
""".encode()

        self.control_handler = self.control_handler_controller.create(
            "Test", "test", self.defect_type
        )

    def test_validate(self):
        self.assertTrue(self.control_handler.validate_source(self.source))

    def test_create_destroy(self):
        self.control_handler.publish_source(self.source)
        self.control_handler.destroy_source()

    def test_materialize(self):
        self.control_handler.publish_source(self.source)
        self.control_handler.materialize_source()
        self.control_handler.destroy_source()


class ControlZoneDatabaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.accessor_controller = AccessorController(self.session)
        self.template_controller = TemplateController(self.session)
        self.control_zone_controller = ControlZoneController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)
        self.template = self.template_controller.create("Test", self.accessor)

    def test_create_delete(self):
        control_zone = self.control_zone_controller.create(
            "Test", self.template, self.accessor
        )
        self.control_zone_controller.delete(control_zone)

    def test_list_nested(self):
        self.control_zone_controller.list_nested()


class ConnectedComponentDatabaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.accessor_controller = AccessorController(self.session)
        self.template_controller = TemplateController(self.session)
        self.control_zone_controller = ControlZoneController(self.session)
        self.connected_component_controller = ConnectedComponentController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)
        self.template = self.template_controller.create("Test", self.accessor)
        self.control_zone = self.control_zone_controller.create(
            self.template, self.accessor
        )

    def test_create_delete(self):
        cc = self.connected_component_controller.create(0, 0, 0, 0, self.control_zone)
        self.connected_component_controller.delete(cc)


class ControlTargetDatabaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.session = Session(engine)
        self.accessor_controller = AccessorController(self.session)
        self.template_controller = TemplateController(self.session)
        self.control_zone_controller = ControlZoneController(self.session)
        self.control_target_controller = ControlTargetController(self.session)
        self.control_handler_controller = ControlHandlerController(self.session)
        self.defect_type_controller = DefectTypeController(self.session)

        self.accessor = self.accessor_controller.retrieve(1)
        self.defect_type = self.defect_type_controller.retrieve(1)
        self.template = self.template_controller.create("Test", self.accessor)
        self.control_zone = self.control_zone_controller.create(
            self.template, self.accessor
        )
        self.control_handler = self.control_handler_controller.create(
            "Test", "test", self.defect_type
        )

    def test_create_delete(self):
        control_target = self.control_target_controller.create(
            self.control_handler, self.control_zone
        )
        self.control_target_controller.delete(control_target)



import io
import logging
from typing import Optional, Tuple, List


from uuid import uuid4

from open_aoi.exceptions import IntegrityError
from open_aoi.mixins import Mixin


logger = logging.getLogger("controller.control_handler")


class Module:
    parameters: List

    @staticmethod
    def validate_specification(specification: dict) -> Tuple[bool, Optional[str]]:
        try:
            assert (
                specification.get("parameters") is not None
            ), "Parameters are missing!"
            assert (
                specification.get("process") is not None
            ), "Process function is missing!"
        except Exception as e:
            return False, str(e)
        return True, None

    def __init__(self, specification: dict, source: bytes) -> None:
        self._spec = specification
        self._source = source
        res, detail = self.validate_specification(specification)
        assert res, detail

        self.parameters = specification["parameters"]
        self.process = specification["process"]

    def process(self):
        raise NotImplementedError()


class ModuleSourceMixin(Mixin):
    handler_blob: Optional[str]
    _bucket_name = "modules"

    @property
    def is_valid(self):
        return getattr(self, "handler_blob", None) is not None

    def publish_source(self, content: bytes) -> str:
        assert getattr(self, "handler_blob", None) is None
        client = self._client

        handler_blob = str(uuid4())

        if not client.bucket_exists(self._bucket_name):
            client.make_bucket(self._bucket_name)

        blob = io.BytesIO(content)
        blob.seek(0)
        client.put_object(self._bucket_name, handler_blob, blob, len(content))

        self.handler_blob = handler_blob

    def materialize_source(self) -> Module:
        assert getattr(self, "handler_blob") is not None
        client = self._client

        if not client.bucket_exists(self._bucket_name):
            raise IntegrityError("Module does not exist")

        obj = client.get_object(self._bucket_name, self.handler_blob)
        source = obj.read()
        obj.close()

        return Module(self._dynamic_import(source), source)

    def destroy_source(self):
        assert getattr(self, "handler_blob") is not None
        client = self._client

        if not client.bucket_exists(self._bucket_name):
            raise IntegrityError("Module does not exist")

        client.remove_object(self._bucket_name, self.handler_blob)
        
        self.handler_blob = None

    @classmethod
    def validate_source(cls, source: bytes) -> Tuple[bool, Optional[str]]:
        try:
            source = cls._dynamic_import(source)
            Module.validate_specification(source)
        except Exception as e:
            return False, str(e)
        return True, None

    @staticmethod
    def _dynamic_import(source: bytes):
        """
        Import dynamically generated code as a module.
        """

        ctx = {}
        exec(source.decode(), ctx, ctx)
        return ctx
